import time, os, warnings, pickle
from scipy.optimize import curve_fit, minimize
from autograd import grad, hessian
import autograd.numpy as np
from utils import kalman_filter, zeus_mcmc, lognorm_pdf, \
				  non_Gauss_1D_filtering, diag_gaussian_2d_pdf, \
				  nearest_PD
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


class state_space_model():
	"""
	Defines the state space model matrices and vectors for FRET estimation
	"""
	
	def __init__(self, analyze_FRET_data_inst):
		"""
		Constructor for state space model.
		"""
		
		self.a = analyze_FRET_data_inst
		
		self._R = np.zeros((self.a.num_IDD, 2, 2))
		self._R[:, 0, 0] = self.a.IDD_noise_sd**2
		self._R[:, 1, 1] = self.a.IDA_noise_sd**2
		
		self.crstlk_a = self.a.data_dict['crstlk']['a']
		self.crstlk_d = self.a.data_dict['crstlk']['d']
		
		self._y = np.vstack((self.a.IDD, self.a.IDA)).T
		
		# Initial statistics of Gaussian distribution
		self.P_0 = np.array([[self.a.anl_params.P0_x1, 0], 
							 [0, self.a.anl_params.P0_chi]])
		self.m_0 = np.array([1, self.a.anl_params.m0_chi])
		
	def Q(self, params):
		"""
		Returns process noise matrix assuming Gaussian process noise on chi.
		"""
		
		sigma_chi_idx = 2 + self.a.num_params_f_D + self.a.num_params_f_A
		_Q = np.array([[0, 0], [0, params[sigma_chi_idx]**2.0]]) 
		return _Q
				
	def H(self, params):
		"""
		Returns measurement matrix at all timepoints, for given parameters
		"""
		
		# Crosstalk parameters
		G = self.a.data_dict['crstlk']['G']
		
		# Evaluate functions at ALL timepoints at once
		f_D_params = params[2: 2 + self.a.num_params_f_D]
		f_A_params = params[2 + self.a.num_params_f_D: 
							2 + self.a.num_params_f_D + self.a.num_params_f_A]
		f_D = self.a.m.f_D.func(self.a.tDD, f_D_params)
		f_A = self.a.m.f_A.func(self.a.tDD, f_A_params)
		
		DT = params[0]
		AT = params[1]
		_H00 = DT*f_D
		_H01 = -f_D*f_A
		_H10 = self.crstlk_d*DT*f_D + self.crstlk_a*AT*f_A
		_H11 = (G - self.crstlk_d)*f_D*f_A
		_H = np.array([[_H00, _H01], [_H10, _H11]])
		_H = np.moveaxis(_H, [0, 1, 2], [1, 2, 0])
		
		return _H
	
	def R(self):
		"""
		Returns measurement covariance at all times, as a function of params
		"""
		
		return self._R
		
	def y(self):
		"""
		Returns data vectors at all times
		"""
		
		return self._y
		

class param_est():
	"""
	Class to run all parameter estimation routines
	"""
	
	def __init__(self, analyze_FRET_data_inst):
		"""
		Constructor takes `analyze_FRET_data' instance and inherits its attrs
		"""
	
		self.a = analyze_FRET_data_inst
		out_dir = self.a.res_dir + '/param_dists'
		if not os.path.exists(out_dir):
			os.makedirs(out_dir)
		
	def plot_initial_est_bleach(self, f_D_prior_loc, f_A_prior_loc):
		"""
		Plotting routines for initial bleaching curve estimation
		"""
	
		fig = plt.figure(figsize=(8, 3))
		ax1 = plt.subplot(121)
		ax1.set_title(r'$I_{DD}$', fontsize=15)
		ax1.plot(self.a.tDD, self.a.IDD, color='0.5', lw=1, 
				 label=r'Measured $I_{DD}$')
		ax1.plot(self.a.tDD, f_D_prior_loc[-1]*self.a.m.f_D.func(self.a.tDD, 
				 f_D_prior_loc[:-1]), color='r', lw=2, label='Fit')
		ax1.tick_params(which='both', labelsize=10)
		ax1.legend()
		ax1.set_xlabel('Time (s)')
		ax1.set_ylabel('Intensity')
		
		ax2 = plt.subplot(122)
		ax2.set_title(r'$I_{AA}$', fontsize=15)
		ax2.plot(self.a.tAA, self.a.IAA, color='0.5', lw=1, 
				 label=r'Measured $I_{AA}$')
		ax2.plot(self.a.tAA, f_A_prior_loc[-1]*self.a.m.f_A.func(self.a.tAA, 
				 f_A_prior_loc[:-1]), color='r', lw=2, label='Fit')
		ax2.tick_params(which='both', labelsize=10)
		ax2.legend()
		ax2.set_xlabel('Time (s)')
		
		plt.tight_layout()
		out_dir = self.a.res_dir + '/param_dists'
		plt.savefig('%s/initial_est_bleach_trend.png' % out_dir, dpi=300)
		plt.close('all')
		
	def calc_prior_modes(self):
		"""
		Finds the modes of the prior distributions by naive curve fitting of
		the bleaching curves. This fixes the priors for the paramters 
		defining f_A and f_D
		"""
	
		# Fit donor bleaching curve to get mode of prior dist. 
		# Add the initial IDD value to be estimated (gives scaling constant)
		f_D_params_init = self.a.m.f_D.p_init + [self.a.IDD[0]]
		f_A_params_init = self.a.m.f_A.p_init + [self.a.IAA[0]]
		
		# Bounds for f_D and f_A parameters
		f_D_lo_bnd = self.a.m.f_D.p_lo_bnd + [0]
		f_D_hi_bnd = self.a.m.f_D.p_hi_bnd + [np.inf]
		f_A_lo_bnd = self.a.m.f_A.p_lo_bnd + [0]
		f_A_hi_bnd = self.a.m.f_A.p_hi_bnd + [np.inf]
		
		def f_D_bleach(t, *params):
			return params[-1]*self.a.m.f_D.func(t, params[:-1])
		def f_A_bleach(t, *params):
			return params[-1]*self.a.m.f_A.func(t, params[:-1])
		f_D_prior_loc = curve_fit(f_D_bleach, self.a.tDD, self.a.IDD, 
								  p0=f_D_params_init, 
								  bounds=(f_D_lo_bnd, f_D_hi_bnd))[0]
		f_A_prior_loc = curve_fit(f_A_bleach, self.a.tAA, self.a.IAA, 
								  p0=f_A_params_init, 
								  bounds=(f_A_lo_bnd, f_A_hi_bnd))[0]
		
		# Just check that the fit lines up to the data trend
		self.plot_initial_est_bleach(f_D_prior_loc, f_A_prior_loc)
			
		# Aggregate modes of all param priors. Order is: DT, AT, [f_D params], 
		# [f_A params], [process noise params].
		# Note that DT is defined by fitting (IDD[0] and IDA[0]), rather
		# than taking the coefficient of the curve fit
		prior_modes = []
		DT = (self.a.IDA[0] + 
		  (self.a.data_dict['crstlk']['G'] - self.a.data_dict['crstlk']['d'])
		  *f_D_prior_loc[-1] - self.a.data_dict['crstlk']['a']
		  *f_A_prior_loc[-1])/self.a.data_dict['crstlk']['G']
		AT = f_A_prior_loc[-1]
		prior_modes.append(DT)
		prior_modes.append(AT)
		prior_modes.extend(f_D_prior_loc[:-1])
		prior_modes.extend(f_A_prior_loc[:-1])
		if self.a.anl_params.process_noise == 'Gaussian':
			prior_modes.append(self.a.anl_params.sigma_chi_prior_mode) 
		elif self.a.anl_params.process_noise == 'Non-Gaussian':
			for mode in self.a.anl_params.Q_tilde_prior_modes:
				prior_modes.append(mode)
		assert len(prior_modes) == self.a.num_params, "Length of prior modes "\
		  "(%s) must be number of total parameters = %s"\
		  % (len(prior_modes), self.a.num_params)
		
		return np.array(prior_modes)
		
	def gen_prior_dist(self):
		"""
		Fixes prior distributions by finding the mode of the hyperparameter
		distributions, then setting these hyperparameters as fixed.
		"""
		
		# Sets just the functional forms of the prior distributions. pdfs
		# are a function of the hypeparameters still, though.
		prior_dist_funcs = self.a.m.prior_dict.func(self.a.anl_params)
		
		# Maximize prior distributions over their hyperparameters. 
		self.prior_modes = self.calc_prior_modes()
		
		# Now, set hyperparameter to optimum. Prior distributions are now fixed.
		# Order is same as usd in self.calc_prior_modes: DT, AT, [f_D params], 
		# [f_A params], [process noise params]
		self.priors = [prior_dist_funcs[self.a.param_names[i]]
						(self.prior_modes[i]) for i in range(self.a.num_params)]
		
	def calc_obs_likelihoods(self, p, chis):
		"""
		Calculate an observation likelihood at each timepoint, for an array
		of chis. Used for non-Gaussian proccess noise in Bayes filtering. 
		"""
		
		int_N = len(chis)
		H = self.ssm.H(p)
		R = self.ssm.R()
		y = self.ssm.y()
		LHs = []
		for i in range(self.a.num_IDD):
			mean1 = H[i, 0, 0]*np.ones(int_N) + H[i, 0, 1]*chis
			mean2 = H[i, 1, 0]*np.ones(int_N) + H[i, 1, 1]*chis
			means = np.vstack((mean1, mean2))
			covs = np.array([R[i] for j in range(int_N)])
			covs = np.moveaxis(covs, [0, 1, 2], [2, 0, 1])
			
			# Copy observations to all chi values to vectorize computation
			ys = np.array([y[i] for j in range(int_N)]).T
			LHs.append(diag_gaussian_2d_pdf(ys, means, covs))
		LHs = np.array(LHs)
		
		return LHs

	def calc_non_gauss_chi_range(self):
		""" 
		Adaptively calculate the range of chis for the integration for
		Non-Gaussian filtering
		"""
	
		a = self.a.data_dict['crstlk']['a']
		d = self.a.data_dict['crstlk']['d']
		G = self.a.data_dict['crstlk']['G']
		DT = self.prior_modes[0]
		AT = self.prior_modes[1]
		idx_lo = 2 + self.a.num_params_f_D
		idx_hi = idx_lo + self.a.num_params_f_A
		f_A_params = self.prior_modes[idx_lo:idx_hi]

		# Calculate estimated chi values for whole time series
		IAA_est = AT*self.a.m.f_A.func(self.a.tDD, f_A_params)
		Fc = np.max(self.a.IDA - d*self.a.IDD - a*IAA_est, 0)
		R = Fc/self.a.IDD
		E_corr = R/(R + G)*(IAA_est[0]/IAA_est)
		chi_est =  DT*E_corr

		chi_med = np.median(chi_est)
		chi_range =  np.percentile(chi_est, 99) - np.percentile(chi_est, 1)
		max = chi_med + (self.a.anl_params.chi_interval_width/2)*chi_range
		min = chi_med - (self.a.anl_params.chi_interval_width/2)*chi_range
		
		chis = np.linspace(min, max, self.a.anl_params.chi_num_of_subintervals)
		
		return chis

	def neg_log_posterior(self, log_p):
		""" 
		Calculates the negative log of the posterior of the parameters given 
		the data.
		"""
		
		p = np.exp(log_p)
		
		# log[prob(p)]
		neg_log_prior = -np.sum([np.log(f(p)) for f, p in zip(self.priors, p)])
		
		AT = p[1]
		f_A_params = p[2 + self.a.num_params_f_D: 
					   2 + self.a.num_params_f_D + self.a.num_params_f_A]
		
		# Technically the commented term term is in the posterior, but does 
		# not affect the optimal estimate since it's a scalar
		#neg_log_IAA = 0.5*np.sum(np.log(self.a.IAA_noise_sd**2))
		neg_log_IAA = 0.5*np.sum((self.a.IAA - AT*self.a.m.f_A.func(
								  self.a.tAA, f_A_params))**2/
								  self.a.IAA_noise_sd**2)
		
		# log[prob(y == IDD, IDA | p)]
		if self.a.anl_params.process_noise == 'Gaussian':
			_, _, _, _, S, v = kalman_filter(self.ssm.Q(p), self.ssm.H(p),
											 self.ssm.R(), self.ssm.y(), 
											 self.ssm.m_0, self.ssm.P_0, 
											 self.a.tDD)
		
			# Perhaps there is a cleaner way, but at least it's vectorized
			S_inv = np.linalg.inv(S)
			vS1 = np.sum(S_inv[:, 0, :]*v, axis=-1)
			vS2 = np.sum(S_inv[:, 1, :]*v, axis=-1)
			vS = np.vstack((vS1, vS2)).T
			vSv = np.sum(vS*v, axis=-1)
			neg_log_y = 0.5*(np.sum(vSv + np.log(np.linalg.det(2*np.pi*S))))
			
		elif self.a.anl_params.process_noise == 'Non-Gaussian':
			
			# Define state transition pdf; here we assume deterministic
			# dynamics are trivial, x --> x + q; q is defined by q_tilde
			idx_lo = 2 + self.a.num_params_f_D + self.a.num_params_f_A
			idx_hi = idx_lo + self.a.num_params_Q_tilde
			Q = self.a.m.Q_tilde.pdf(p[idx_lo:idx_hi])
			
			m_0 = self.a.anl_params.m0_chi
			P_0 = self.a.anl_params.P0_chi
			
			chis = self.calc_non_gauss_chi_range()
			
			# Observation likelihoods
			LHs = self.calc_obs_likelihoods(p, chis)
			
			# Filter and smooth by integrating over chi at each timestep
			dist = non_Gauss_1D_filtering(chis, Q, LHs, m_0, P_0, self.a.tDD)
			
			dx, p1 = dist['dx'], dist['pred_probs']
			neg_log_y = -np.sum(np.log(np.sum(dx*p1*LHs, axis=-1)), axis=0)
		
		log_post = neg_log_prior + neg_log_IAA + neg_log_y
		print (log_post)
		return  log_post
		
	def log_posterior(self, log_p):
		"""
		Used to maximize the posterior for MCMC
		"""
		return -self.neg_log_posterior(log_p)
		
	def gen_post_dist_laplace_approx(self):
		"""
		Use Laplace approximation to generate the posterior by maximimizing 
		the log likelihood, then getting the Hessian to approximate as a 
		normal distribution.
		"""
		
		print ('Optimizing log posterior to get mode for Laplace approx..')
		timenow = time.time()
		with warnings.catch_warnings():
			warnings.simplefilter('ignore')
			
			# Impose bounds for the bleaching parameters as provided by user
			# Let's keep everything between 1e-8 and 1e8 to avoid overflow
			nmax = 1e8
			nmin = 1e-8
			bounds = np.array([[np.log(nmin), np.log(nmax)]]*self.a.num_params)
			for i in range(self.a.num_params_f_D):
				_i = i + 2
				bounds[_i][0] = np.log(max(self.a.m.f_D.p_lo_bnd[i], nmin))
				bounds[_i][1] = np.log(min(self.a.m.f_D.p_hi_bnd[i], nmax))
			for i in range(self.a.num_params_f_A):
				_i = i + self.a.num_params_f_D + 2
				bounds[_i][0] = np.log(max(self.a.m.f_A.p_lo_bnd[i], nmin))
				bounds[_i][1] = np.log(min(self.a.m.f_A.p_hi_bnd[i], nmax))
			
			# Minimize using sequential least squares programming
			res = minimize(self.neg_log_posterior, np.log(self.prior_modes),
						   bounds=bounds, method='SLSQP',
						   options=self.a.anl_params.MAP_opt_options)
			
		# Hessian of log posterior is with respect to *log* of parameters.
		self.MAP_log_hess = hessian(self.neg_log_posterior)(res.x)
		self.MAP_inv_log_hess = np.linalg.inv(self.MAP_log_hess)
		eigs = np.linalg.eig(self.MAP_inv_log_hess)[0]
		if np.any(eigs < 0):
			self.MAP_inv_log_hess = nearest_PD(self.MAP_inv_log_hess)
			eigs = np.linalg.eig(self.MAP_inv_log_hess)[0]
		self.MAP_log_p = res.x
		self.BIC = 2*res.fun + self.a.num_params*np.log(self.a.num_IDD 
														+ self.a.num_IDA 
														+ self.a.num_IAA)
		print ('Function value: ', res.fun, 'BIC:', self.BIC,  
			   'Opt success:', res.success, '\n', 'Optimum:', np.exp(res.x))
		print ('Covariance matrix eigs', eigs)
		print ('Time elapsed:', time.time() - timenow)
		
	def plot_prior_and_post_dists(self):
		"""
		Plots prior and posterior distributions in both linear and log.
		"""
		
		for iP in range(self.a.num_params):
			
			fig = plt.figure(figsize=(8, 3))
			ax1 = plt.subplot(121)
			ax2 = plt.subplot(122)
			
			# Plot priors on log scale
			log_mid = np.log(self.prior_modes[iP])/np.log(10)
			x = np.logspace(log_mid - 4, log_mid + 4, 100000)
			y = self.priors[iP](x)
			idx_rng = y > 0.1*max(y)
			ax1.plot(x[idx_rng], y[idx_rng], color='dodgerblue', 
					 lw=2, label='Prior')
			ax1.set_xlim(min(x[idx_rng]), max(x[idx_rng]))
			ax1.set_ylim(0, 2*max(y))
			
			# Plot posterior on log scale
			if self.a.anl_params.do_MCMC:
				mode = np.median(self.post_p_samples[:, iP])
				ax1.hist(self.post_p_samples[:, iP], bins=100, density=True, 
						 color='r', label='Posterior')
			else:
				std = self.MAP_inv_log_hess[iP, iP]**0.5
				mode = np.exp(self.MAP_log_p[iP])
				y = lognorm_pdf(std).pdf(mode)(x)
				ax1.plot(x, y, color='r', label='Posterior')
			ax1.axvline(mode, ls='--', color='r', label='MAP')
			
			# Repeat for linear scale
			mid = self.prior_modes[iP]
			x = np.linspace(mid/1e3, mid*1e3, 100000)
			y = self.priors[iP](x)
			idx_rng = y > 0.1*max(y)
			ax2.plot(x[idx_rng], y[idx_rng], color='dodgerblue', 
					 lw=2, label='Prior')
			ax2.set_xlim(min(x[idx_rng]), max(x[idx_rng]))
			ax2.set_ylim(0, 2*max(y))
			if self.a.anl_params.do_MCMC:
				mode = np.median(self.post_p_samples[:, iP])
				ax2.hist(self.post_p_samples[:, iP], bins=100, density=True, 
						 color='r', label='Posterior')
			else:			
				y = lognorm_pdf(std).pdf(mode)(x)
				ax2.plot(x, y, color='r', label='Posterior')
			ax2.axvline(mode, ls='--', color='r', label='MAP')
			
			name = self.a.param_names[iP]
			ax1.set_title(name)
			ax1.tick_params(which='both', labelsize=10)
			ax1.set_xscale('log')
			ax1.set_ylabel('pdf')
			ax1.legend()
			ax2.set_title(name)
			ax2.tick_params(which='both', labelsize=10)
			ax2.set_xlabel(name)
			ax2.legend()
			plt.tight_layout()
			out_dir = self.a.res_dir + '/param_dists'
			plt.savefig('%s/%s.png' % (out_dir, name), dpi=300)
		
	def sample_from_post_dist(self):
		"""
		Gets samples from posterior of parameters by doing either Laplace 
		approximation or MCMC. These are then used to estimate the states.
		"""
		
		self.ssm = state_space_model(self.a)
		
		vars = ['post_p_samples']
		if self.a.anl_params.do_MCMC:
			log_samples = zeus_mcmc(self.log_posterior, 
									np.log(self.prior_modes), 
									self.a.anl_params.MCMC_num_workers, 
									self.a.anl_params.MCMC_num_samples, 
									burnin=self.a.anl_params.MCMC_burnin, 
									all_cores=self.a.anl_params.MCMC_all_cores,
									d_init=self.a.anl_params.MCMC_d_init)
			
		else:
			self.gen_post_dist_laplace_approx()
			log_samples = np.random.multivariate_normal(
						    self.MAP_log_p, 
							self.MAP_inv_log_hess, 
							self.a.anl_params.MAP_num_samples)
			vars += ['MAP_log_p', 'MAP_log_hess', 'MAP_inv_log_hess', 'BIC']
		self.post_p_samples = np.exp(log_samples)
		
		# Save results and plot the parameter distributions
		data = dict()
		for var in vars:
			exec("data['%s'] = self.%s" % (var, var))
		out_dir = self.a.res_dir + '/param_dists'
		if self.a.anl_params.do_MCMC:
			filename = '%s/posterior_data_MCMC.pkl' % out_dir
		else:
			filename = '%s/posterior_data.pkl' % out_dir
		pickle.dump(data, open(filename, 'wb'))
		self.plot_prior_and_post_dists()	
		


import numpy as np
from numpy import ma

import matplotlib.pyplot as plt
import matplotlib.patches as patches

from skimage import morphology
from skimage import exposure
from skimage import measure
from skimage import filters

from scipy import ndimage as ndi
from scipy import stats

from ..utils import masking
from ..utils import plot


class wt_vs_mut():
    def __init__(self, wt_img:np.ndarray, mut_img:np.ndarray,
                 proc_mask:np.ndarray, narrow_proc_mask:np.ndarray,
                 **kwargs):
        """ Class is designed to create differential images and insertion masks
        for two separate image time series (for instance, wild-type NCS and mutant NCS,
        or two distinct NCSs). The up_mask_calc method is responsible for detecting
        insertion regions.

        Suitable for image series with __single__ application/stimulation only.

        The class attributes contain various types of insertion masks
        and corresponding profiles for these masks (raw intensity and ΔF/F profiles).

        __Requires wf_x2_m2 instance type as input!__

        __NB: Optimized for long stimuli/applications with strong translocation
        and analysis of colocalization of the insertion sites.
        Methods of this class may not work with image series with weak translocation range.__

        Parameters
        ----------
        wt_img: ndarray [t,x,y]
            wild type images time series
        mut_img: ndarray [t,x,u]
            mutant images time series
        proc_mask: ndarray [x,y]
            cell processes boolean mask, extended
        narrow_proc_mask: ndarray [x,y]
            cell processes boolean mask, unextended

        Attributes
        ----------
        wt_img: ndarray [t,x,y]
            wild type images time series
        mut_img: ndarray [t,x,y]
            mutant images time series
        proc_mask: ndarray [x,y]
            cell processes boolean mask, extended
        narrow_proc_mask: ndarray [x,y]
            cell processes boolean mask, unextended
        wt_up_mask: ndarray [x,y]
            boolean mask of intensity increase regions for wild-type channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        wt_up_label: ndarray [x,y]
            label image of intensity increase regions for wild-type channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        wt_diff_img: ndarray [x,y]
            differential image of intensity changes after stimulation for wild-type channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        mut_up_mask: ndarray [x,y]
            boolean mask of intensity increase regions for mutant/2nd NCS channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        mut_up_label: ndarray [x,y]
            label image of intensity increase regions for mutant/2nd NCS channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        mut_diff_img: ndarray [x,y]
            differential image of intensity changes after stimulation for mutant/2nd NCS channel,
            created with `red_green.wt_ws_mut.up_mask_calc()`
        connected_up_mask: ndarray [x,y]
            regions of `wt_up_mask` which overlay with `mut_up_mask`,
            created with `utils.masking.mask_connection()`
        halo_up_mask: ndarray [x,y]
            `connecter_up_mask` without mutant intensity increase regions
        init_up_mask: ndarray [x,y]
            `connected_up_mask` without `halo_up_mask`
            (only regions of `mut_up_mask` which overlay with `wt_up_mask`)
        wt_conn_arr: ndarray [int, t]
            intensity profiles of wild-type channel for each label element of `connected_up_mask`,
            created with `utils.masking.label_prof_arr()`
        wt_conn_df_arr: ndarray [ΔF/F, t]
            ΔF/F profiles of wild-type channel for each label element of `connected_up_mask`,
            created with `utils.masking.label_prof_arr()`
        wt_halo_arr: ndarray [int, t]
            intensity profiles of wild-type channel for each label element of `halo_up_mask`,
            created with `utils.masking.label_prof_arr()`
        wt_halo_df_arr: ndarray [ΔF/F, t]
            ΔF/F profiles of wild-type channel for each label element of `halo_up_mask`,
            created with `utils.masking.label_prof_arr()`
        wt_init_arr: ndarray [int, t]
            intensity profiles of wild-type channel for each label element of `init_up_mask`,
            created with `utils.masking.label_prof_arr()`
        wt_init_df_arr: ndarray [ΔF/F, t]
            ΔF/F profiles of wild-type channel for each label element of `init_up_mask`,
            created with `utils.masking.label_prof_arr()`
        mut_conn_arr: ndarray [int, t]
            intensity profiles of mutant/2nd NCS channel for each label element of `connected_up_mask`,
            created with `utils.masking.label_prof_arr()`
        mut_conn_df_arr: ndarray [ΔF/F, t]
            ΔF/F profiles of mutant/2nd NCS channel for each label element of `connected_up_mask`,
            created with `utils.masking.label_prof_arr()`
        mut_halo_arr: ndarray [int, t]
            intensity profiles of mutant/2nd NCS channel for each label element of `halo_up_mask`,
            created with `utils.masking.label_prof_arr()`
        mut_halo_df_arr: ndarray [ΔF/F, t]
            ΔF/F profiles of mutant/2nd NCS channel for each label element of `halo_up_mask`,
            created with `utils.masking.label_prof_arr()`
        mut_init_arr: ndarray [int, t]
            intensity profiles of mutant/2nd NCS channel for each label element of `init_up_mask`,
            created with `utils.masking.label_prof_arr()`
        mut_init_df_arr: ndarray [ΔF/F, t]
            ΔF/F profiles of mutant/2nd NCS channel for each label element of `init_up_mask`,
            created with `utils.masking.label_prof_arr()`

        """
        self.wt_img = wt_img
        self.mut_img = mut_img
        self.proc_mask = proc_mask
        self.narrow_proc_mask = narrow_proc_mask

        # WT masking
        self.wt_up_mask, self.wt_up_label, self.wt_diff_img = self.up_mask_calc(self.wt_img,
                                                                                self.proc_mask,
                                                                                **kwargs)

        # mutant masking
        self.mut_up_mask, self.mut_up_label, self.mut_diff_img = self.up_mask_calc(self.mut_img,
                                                                                   self.proc_mask,
                                                                                   **kwargs)

        # masks modification
        self.connected_up_mask, self.connected_up_label = masking.mask_connection(input_master_mask=self.wt_up_mask,
                                                                                  input_minor_mask=self.mut_up_mask)
        self.halo_up_mask = np.copy(self.connected_up_mask)
        self.halo_up_mask[self.mut_up_mask] = 0
        self.halo_up_label = np.copy(self.connected_up_label)
        self.halo_up_label[self.mut_up_mask] = 0

        self.init_up_mask = np.copy(self.connected_up_mask)
        self.init_up_mask[self.halo_up_mask] = 0
        self.init_up_label = np.copy(self.connected_up_label)
        self.init_up_label[self.halo_up_mask] = 0

        # WT ins profiles calc
        self.wt_conn_df_arr, self.wt_conn_arr = masking.label_prof_arr(input_label=self.connected_up_label,
                                                                       input_img_series=self.wt_img)
        self.wt_halo_df_arr, self.wt_halo_arr = masking.label_prof_arr(input_label=self.halo_up_label,
                                                                       input_img_series=self.wt_img)
        self.wt_init_df_arr, self.wt_init_arr = masking.label_prof_arr(input_label=self.init_up_label,
                                                                       input_img_series=self.wt_img)
        
        # # WT trans profiles calc
        # self.wt_conn_rois_trans_arr, self.wt_conn_tot_trans_arr = masking.trans_prof_arr(input_total_mask=self.narrow_proc_mask,
        #                                                                                  input_labels=self.connected_up_label,
        #                                                                                  input_img_series=self.wt_img)
        # self.wt_halo_rois_trans_arr, self.wt_halo_tot_trans_arr = masking.trans_prof_arr(input_total_mask=self.narrow_proc_mask,
        #                                                                                  input_labels=self.halo_up_label,
        #                                                                                  input_img_series=self.wt_img)
        # self.wt_init_rois_trans_arr, self.wt_init_tot_trans_arr = masking.trans_prof_arr(input_total_mask=self.narrow_proc_mask,
        #                                                                                  input_labels=self.init_up_label,
        #                                                                                  input_img_series=self.wt_img)

        # mut profiles calc
        self.mut_conn_df_arr, self.mut_conn_arr = masking.label_prof_arr(input_label=self.connected_up_label,
                                                                         input_img_series=self.mut_img)
        self.mut_halo_df_arr, self.mut_halo_arr = masking.label_prof_arr(input_label=self.halo_up_label,
                                                                         input_img_series=self.mut_img)
        self.mut_init_df_arr, self.mut_init_arr = masking.label_prof_arr(input_label=self.init_up_label,
                                                                         input_img_series=self.mut_img)
        
        # # mut trans profiles calc
        # self.mut_conn_rois_trans_arr, self.mut_conn_tot_trans_arr = masking.trans_prof_arr(input_total_mask=self.narrow_proc_mask,
        #                                                                                  input_labels=self.connected_up_label,
        #                                                                                  input_img_series=self.mut_img)
        # self.mut_halo_rois_trans_arr, self.mut_halo_tot_trans_arr = masking.trans_prof_arr(input_total_mask=self.narrow_proc_mask,
        #                                                                                  input_labels=self.halo_up_label,
        #                                                                                  input_img_series=self.mut_img)
        # self.mut_init_rois_trans_arr, self.mut_init_tot_trans_arr = masking.trans_prof_arr(input_total_mask=self.narrow_proc_mask,
        #                                                                                  input_labels=self.init_up_label,
        #                                                                                  input_img_series=self.mut_img)


    @staticmethod
    def up_mask_calc(input_img_series:np.ndarray, input_img_mask:np.ndarray,
                     sd_tolerance:int=2,
                     base_frames:int=5, stim_start:int=7, stim_win:int=5):
        """ Function for generating an insertion regions mask
        using the differential image method.

        Could be used in stand-alone mode as a static method of the WTvsMut class.

        Parameters
        ----------
        input_img_series: ndarray [t,x,y]
            image time series
        input_img_mask: ndarray [x,y]
            cell region boolean mask
        sd_tolerance: int
            insertion ("up") region detection threshold: the number
            of standard deviations of extracellular noise
            (measured in the area outside of `input_img_mask`)
        base_frames: int
            number of frames from the beginning
            of the image series used to create
            an image of basal fluorescence
        stim_start: int
            index of the frame where stimulation begins
        stim_win: int
            stimulation window, the number of frames following `stim_start` index
            that are used to create an image displaying maximal insertions

        Returns
        -------
        up_mask_filt: ndarray [x,y]
            boolean mask of intensity increase regions
        up_label: ndarray [x,y]
            label image of intensity increase regions  
        img_diff: ndarray [x,y]
            differential image of intensity changes after stimulation

        """
        ref_img_series = filters.gaussian(input_img_series, sigma=1.25, channel_axis=0)

        img_base = np.mean(ref_img_series[:base_frames], axis=0)
        img_max = np.mean(ref_img_series[stim_start:stim_start+stim_win], axis=0)

        img_diff = img_max - img_base
        img_diff = img_diff/np.max(np.abs(img_diff))

        diff_sd = np.std(ma.masked_where(~input_img_mask, img_diff))
        up_mask = img_diff > diff_sd * sd_tolerance

        up_mask_filt = morphology.opening(up_mask, footprint=morphology.disk(2))
        up_mask_filt = morphology.dilation(up_mask_filt, footprint=morphology.disk(1))
        up_label = measure.label(up_mask_filt)

        return up_mask_filt, up_label, img_diff


    def df_mean_prof_pic(self, fsize=(10,10), p_size=15, stim_t=8):
        time_line = np.linspace(0, self.wt_conn_df_arr.shape[1]*2, num=self.wt_conn_df_arr.shape[1])

        arr_stat = lambda x: (np.mean(x, axis=0),  np.std(x, axis=0)/np.sqrt(x.shape[1]))

        wt_conn_df_mean, wt_conn_df_sem = arr_stat(self.wt_conn_df_arr)
        wt_halo_df_mean, wt_halo_df_sem = arr_stat(self.wt_halo_df_arr)
        wt_init_df_mean, wt_init_df_sem = arr_stat(self.wt_init_df_arr)

        mut_conn_df_mean, mut_conn_df_sem = arr_stat(self.mut_conn_df_arr)
        mut_halo_df_mean, mut_halo_df_sem = arr_stat(self.mut_halo_df_arr)
        mut_init_df_mean, mut_init_df_sem = arr_stat(self.mut_init_df_arr)

        y_max = np.vstack([wt_conn_df_mean, wt_halo_df_mean, wt_init_df_mean, \
                           mut_conn_df_mean, mut_halo_df_mean, mut_init_df_mean]).max()
        y_max = y_max + y_max*0.1
        y_min = np.vstack([wt_conn_df_mean, wt_halo_df_mean, wt_init_df_mean, \
                           mut_conn_df_mean, mut_halo_df_mean, mut_init_df_mean]).min()
        y_min = y_min + y_min*0.1

        def two_arr_u_test(arr_1, arr_2):
            p_list = []
            for i in range(arr_1.shape[1]):
                t_1 = arr_1[:,i]
                t_2 = arr_2[:,i]
                amp_u_test = stats.mannwhitneyu(t_1, t_2)
                p_val = amp_u_test[1]

                if p_val > 0.05:
                    p_list.append(' ')
                elif p_val > 0.01:
                    p_list.append('*')
                elif p_val <= 0.01:
                    p_list.append('**')
            return p_list

        plt.figure(figsize=fsize)
        ax0 = plt.subplot(311)
        ax0.set_title('Connected up mask')
        ax0.errorbar(time_line, wt_conn_df_mean,
                     yerr = wt_conn_df_sem,
                     fmt ='-o', color='k', capsize=2, label='WT')
        ax0.errorbar(time_line, mut_conn_df_mean,
                     yerr = mut_conn_df_sem,
                    fmt ='-o', color='r', capsize=2, label='Mut.')
        conn_u_test_p = two_arr_u_test(self.wt_conn_df_arr, self.mut_conn_df_arr)
        for idx in range(len(time_line)):
            ax0.text(x=time_line[idx], y=wt_conn_df_mean[idx]*1.1,
                     s=conn_u_test_p[idx], fontsize=p_size)  
        ax0.axvline(x=stim_t, linestyle=':', color='k', label='Glu application')      
        ax0.hlines(y=0, xmin=0, xmax=time_line.max(), linestyles='--', color='k')
        ax0.set_xlabel('Time, s')
        ax0.set_ylabel('ΔF/F')
        ax0.set_ylim([y_min, y_max])
        ax0.legend()

        ax1 = plt.subplot(312)
        ax1.set_title('Halo up mask')
        ax1.errorbar(time_line, wt_halo_df_mean,
                     yerr = wt_halo_df_sem,
                     fmt ='-o', color='k', capsize=2, label='WT')
        ax1.errorbar(time_line, mut_halo_df_mean,
                     yerr = mut_halo_df_sem,
                    fmt ='-o', color='r', capsize=2, label='Mut.')
        halo_u_test_p = two_arr_u_test(self.wt_halo_df_arr, self.mut_halo_df_arr)
        for idx in range(len(time_line)):
            ax1.text(x=time_line[idx], y=wt_halo_df_mean[idx]*1.1,
                     s=halo_u_test_p[idx], fontsize=p_size)  
        ax1.axvline(x=stim_t, linestyle=':', color='k', label='Glu application')      
        ax1.hlines(y=0, xmin=0, xmax=time_line.max(), linestyles='--', color='k')
        ax1.set_xlabel('Time, s')
        ax1.set_ylabel('ΔF/F')
        ax1.set_ylim([y_min, y_max])
        ax1.legend()

        ax2 = plt.subplot(313)
        ax2.set_title('Init up mask')
        ax2.errorbar(time_line, wt_init_df_mean,
                     yerr = wt_init_df_sem,
                     fmt ='-o', color='k', capsize=2, label='WT')
        ax2.errorbar(time_line, mut_init_df_mean,
                     yerr = mut_init_df_sem,
                    fmt ='-o', color='r', capsize=2, label='Mut.')
        init_u_test_p = two_arr_u_test(self.wt_init_df_arr, self.mut_init_df_arr)
        for idx in range(len(time_line)):
            ax2.text(x=time_line[idx], y=wt_init_df_mean[idx]*1.1,
                     s=init_u_test_p[idx], fontsize=p_size)  
        ax2.axvline(x=stim_t, linestyle=':', color='k', label='Glu application')      
        ax2.hlines(y=0, xmin=0, xmax=time_line.max(), linestyles='--', color='k')
        ax2.set_xlabel('Time, s')
        ax2.set_ylabel('ΔF/F')
        ax2.set_ylim([y_min, y_max])
        ax2.legend()

        plt.tight_layout()
        plt.show()
 

    def diff_img_pic(self):
        cell_contour = measure.find_contours(self.proc_mask, level=0.5)

        plt.figure(figsize=(15,15))
        ax0 = plt.subplot(121)
        ax0.set_title(f'Differential img, WT ({np.max(self.wt_up_label)} ROIs)')
        ax0.imshow(self.wt_diff_img, cmap=plot.CMaps().cmap_red_green, vmax=1, vmin=-1)
        for ce_c in cell_contour:
            ax0.plot(ce_c[:, 1], ce_c[:, 0], linewidth=1, color='w')
        ax0.axis('off')

        ax1 = plt.subplot(122)
        ax1.set_title(f'Differential img, mutant ({np.max(self.mut_up_label)} ROIs)')
        ax1.imshow(self.mut_diff_img, cmap=plot.CMaps().cmap_red_green, vmax=1, vmin=-1)
        for ce_c in cell_contour:
            ax1.plot(ce_c[:, 1], ce_c[:, 0], linewidth=1, color='w')
        ax1.axis('off')

        plt.tight_layout()
        plt.show()


    def mask_diff_pic(self):
        up_mask_diff = plot.toRGB(r_img=self.wt_up_mask,
                                  g_img=self.mut_up_mask,
                                  b_img=np.zeros_like(self.wt_up_mask))
        con_v_init_diff = plot.toRGB(r_img=self.connected_up_mask,
                                     g_img=self.init_up_mask,
                                     b_img=np.zeros_like(self.connected_up_mask))

        plt.figure(figsize=(15,15))
        ax0 = plt.subplot(121)
        ax0.set_title('WT up mask (red) vs. mut. up mask (green)')
        ax0.imshow(up_mask_diff)
        ax0.axis('off')

        ax1 = plt.subplot(122)
        ax1.set_title('Connected mask (red) vs. init mask (green)')
        ax1.imshow(con_v_init_diff)
        ax1.axis('off')

        plt.tight_layout()
        plt.show()


    def mask_set_pic(self):
        plt.figure(figsize=(15,15))
        ax0 = plt.subplot(131)
        ax0.set_title('Connected up mask elements')
        ax0.imshow(self.connected_up_label, cmap='jet')
        ax0.axis('off')

        ax1 = plt.subplot(132)
        ax1.set_title('Halo up mask elements')
        ax1.imshow(self.halo_up_label, cmap='jet')
        ax1.axis('off')

        ax2 = plt.subplot(133)
        ax2.set_title('Init up mask elements')
        ax2.imshow(self.init_up_label, cmap='jet')
        ax2.axis('off')

        plt.tight_layout()
        plt.show()


    def connection_contour_pic(self):
        plt.figure(figsize=(10,10))
        plt.imshow(plot.toRGB(self.connected_up_mask, self.init_up_mask, np.zeros_like(self.wt_up_mask)))
        plt.axis('off')
        plt.title('Up mask overlay (red-WT, green-mutant)')
        plt.tight_layout()
        plt.show()


        cell_contour = measure.find_contours(self.proc_mask, level=0.5)
        conn_contour = measure.find_contours(self.connected_up_mask, level=0.5)
        init_contour = measure.find_contours(self.init_up_mask, level=0.5)

        plt.figure(figsize=(20,20))
        plt.imshow(np.mean(self.wt_img, axis=0) *-1, cmap='Greys')
        for ce_c in cell_contour:
            plt.plot(ce_c[:, 1], ce_c[:, 0], linewidth=1, color='w')
        for co_c in conn_contour:
            plt.plot(co_c[:, 1], co_c[:, 0], linewidth=1, color='r', alpha=.75)
        for in_c in init_contour:
            plt.plot(in_c[:, 1], in_c[:, 0], linewidth=1, color='g', alpha=.75)
        
        plt.axis('off')
        plt.tight_layout()
        plt.show()


class anl_params():

	def __init__(a):
	
		a.process_noise = 'Non-Gaussian'
		a.state_est_all_cores = True
		
		if a.process_noise == 'Non-Gaussian':
			
			# Distr of hyperparameter nu for Student's t (lognormal prior)
			# There are 2 parameters here, each with the mode and fold-change 
			# defining the lognormal distribution of their prior distribution
			# sigma and nu, respectively
			a.Q_tilde_prior_modes = [100, 1]
			a.Q_tilde_prior_FC = [30, 10]
			
			# Set to 'SMC' or 'integrate'. 
			a.state_est_method = 'integrate'
			
			# Parameters to do manual integration for chi pdf
			a.chi_interval_width = 8
			a.chi_num_of_subintervals = 400
			
			# Parameters for sequential Monte Carlo (paticle filter)
			a.num_SMC_samples = 300
			
		else:
			
			# Params defining the prior for the sigma of chi's process noise
			# i.e. Q in the SSM is [[0, 0], [0, sigma_chi^2]], where sigma_chi
			# is Bayesian updated from a lognormal prior defined by:
			a.sigma_chi_prior_FC = 30
			a.sigma_chi_prior_mode = 100

		# Use MCMC to sample from posterior? If not, use Laplace approximation
		# Workers needs to be minimum 4 (and must be even) or it will fail.
		# For ensemble slice sampling to work, it is suggested to use 
		# 2*num_params for the number of walkers. However, this is not required.
		# d_init is the size of the ball, in logspace, over which the 
		# workers are initialized. 0.1 ~ 10% of the value seems appropriate.
		# This value must be positive and greater than 0 or sampling will fail.
		a.do_MCMC = False
		if a.do_MCMC:
			a.MCMC_all_cores = True
			a.MCMC_d_init = 0.1
			a.MCMC_num_workers = 14
			a.MCMC_num_samples = 3000
			a.MCMC_burnin = 300
		else:
			a.MAP_num_samples = 1000
			a.MAP_opt_options = {'maxiter': 10000, 'ftol': 1e-8}
			
		# Options for the optimization for curve fitting of bleaching functions
		a.curve_fit_options = dict()
		a.curve_fit_options['ftol'] = 1e-6
		a.curve_fit_options['maxfun'] = 1e5
		a.curve_fit_options['maxiter'] = 4000
		
		# Priors on the state variables for filtering
		# m = [mean of constant variable, mean of chi]
		# == [1, m0_chi]
		# P = [[var of constant variable, 0], [0, variance of chi]]
		# == [[P0_x1, 0], [0, P0_chi]]
		# First state variable is just =1, up to this level of process noise
		a.P0_x1 = 1e-10

		# chi statistics at time 0 -- mean and std of chi estimate at time 0
		a.m0_chi = 0 
		a.P0_chi = 1e6


import autograd.numpy as np
from utils import norm_pdf, lognorm_pdf, standard_uniform_pdf
from autograd.scipy.special import gamma
from scipy.stats import t

class model():
	
	class f_D():

		def __init__(self):
		
			self.p_name = ['tau_D1', 'tau_D2', 'delta_D']
			self.p_init = [100, 100, 0.5]
			self.p_lo_bnd = [0, 0, 0]
			self.p_hi_bnd = [np.inf, np.inf, 1]
		
		def func(self, t, params):
	
			return params[2]*np.exp(-t/params[0]) \
				   + (1 - params[2])*np.exp(-t/params[1])
	
	class f_A():

		def __init__(self):
		
			self.p_name = ['tau_A1', 'tau_A2', 'delta_A']
			self.p_init = [100, 100, 0.5]
			self.p_lo_bnd = [0, 0, 0]
			self.p_hi_bnd = [np.inf, np.inf, 1]
			
		def func(self, t, params):
			
			return params[2]*np.exp(-t/params[0]) \
				   + (1 - params[2])*np.exp(-t/params[1])
		
	class Q_tilde():
		
		def __init__(self):
		
			self.p_name = ['sigma_chi', 'nu']

		def _f(self, x):
			
			val = 1/self.sd*gamma(self.nu/2 + 0.5)/np.sqrt(np.pi*self.nu)\
				  /gamma(self.nu/2)*(1 + (x/self.sd)**2/self.nu)\
				  **(-self.nu/2 - 0.5)
			
			return val

		def sample(self, params, size=1):
			
			sd, nu = params
			
			return t.rvs(df=nu, scale=sd, size=size)

		def pdf(self, params):
			
			self.sd, self.nu = params
			
			return self._f
		
		
	class prior_dict(object):
		"""
		Define functional form for prior distributions for D_T, A_T, but these
		still have a hyperparameter (location) to be fixed.
		"""
	
		def func(self, anl_params):
		
			pdfs_dict = dict()
			
			# Note that the ordering here is irrelevant
			# DT, AT
			pdfs_dict['DT'] = lognorm_pdf(sigma=np.log(2)).pdf
			pdfs_dict['AT'] = lognorm_pdf(sigma=np.log(1.1)).pdf
		
			# f_D params
			pdfs_dict['tau_D1'] =  lognorm_pdf(sigma=np.log(2)).pdf
			pdfs_dict['tau_D2'] = lognorm_pdf(sigma=np.log(2)).pdf
			pdfs_dict['delta_D'] = standard_uniform_pdf().pdf
			
			# f_A params
			pdfs_dict['tau_A1'] =  lognorm_pdf(sigma=np.log(1.1)).pdf
			pdfs_dict['tau_A2'] =  lognorm_pdf(sigma=np.log(1.1)).pdf
			pdfs_dict['delta_A'] =  standard_uniform_pdf().pdf
			
			# This iterates through all hyperparameters of Q_tilde, using 
			# parameter names defined in its constructor
			for i in range(len(anl_params.Q_tilde_prior_modes)):
				FC = anl_params.Q_tilde_prior_FC[i]
				mode = anl_params.Q_tilde_prior_modes[i]
				lg = lognorm_pdf(sigma=np.log(FC), fixed_mode=mode)
				pdfs_dict[model().Q_tilde().p_name[i]] = lg.pdf
			
			return pdfs_dict


"""
Functions for plotting, image creation, cmaps

"""

import numpy as np
from numpy import ma

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.animation as anm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import colors
from matplotlib.colors import LinearSegmentedColormap

from scipy import stats


class CMaps():
    def __init__(self):
        """
        Class for custom colormaps storage.
        To use colormaps with `matplotlib.imshow` just call `plot.CMaps().cmap_name`.

        Attributes
        ----------
        cmap_cyan: matplotlib cmap
            cyan colormap
        cmap_yellow: matplotlib cmap
            yellow colormap
        cmap_red_green: matplotlib cmap
            red-green colormap (low-green, zero-black, high-red),
            especially suitable for differential images

        """
        # CFP cmap
        dict_cyan = {'red':(
                    (0.0, 0.0, 0.0),
                    (1.0, 0.0, 0.0)),
                    'blue':(
                    (0.0, 0.0, 0.0),
                    (1.0, 1.0, 1.0)),
                    'green':(
                    (0.0, 0.0, 0.0),
                    (1.0, 1.0, 1.0))}
        self.cmap_cyan = LinearSegmentedColormap('Cyan', dict_cyan)

        # YFP cmap
        dict_yellow = {'red':(
                    (0.0, 0.0, 0.0),
                    (1.0, 1.0, 1.0)),
                    'blue':(
                    (0.0, 0.0, 0.0),
                    (1.0, 0.0, 0.0)),
                    'green':(
                    (0.0, 0.0, 0.0),
                    (1.0, 1.0, 1.0))}
        self.cmap_yellow = LinearSegmentedColormap('Yellow', dict_yellow)

        # red-green cmap creation
        dict_red_green = {'red':(
                        (0.0, 0.0, 0.0),
                        (0.5, 0.0, 0.0),
                        (0.55, 0.3, 0.7),
                        (1.0, 1.0, 1.0)),
                        'blue':(
                        (0.0, 0.0, 0.0),
                        (1.0, 0.0, 0.0)),
                        'green':(
                        (0.0, 1.0, 1.0),
                        (0.45, 0.7, 0.3),
                        (0.5, 0.0, 0.0),
                        (1.0, 0.0, 0.0))}
        self.cmap_red_green = LinearSegmentedColormap('RedGreen', dict_red_green)


        @staticmethod
        def plot_linearmap(cdict):
            newcmp = LinearSegmentedColormap('testCmap', segmentdata=cdict, N=256)
            rgba = newcmp(np.linspace(0, 1, 256))
            fig, ax = plt.subplots(figsize=(4, 3), constrained_layout=True)
            col = ['r', 'g', 'b']
            for xx in [0.25, 0.5, 0.75]:
                ax.axvline(xx, color='0.7', linestyle='--')
            for i in range(3):
                ax.plot(np.arange(256)/256, rgba[:, i], color=col[i])
            ax.set_xlabel('index')
            ax.set_ylabel('RGB')
            plt.show()


def toRGB(r_img:np.ndarray, g_img:np.ndarray, b_img:np.ndarray):
    """ Convert three 2d arrays to RGB (input arrays must have same size).

    Parameters
    ----------
    r_img: ndarray [x,y]
        2D array for red chennel
    g_img: ndarray [x,y]
        2D array for green chennel
    b_img: ndarray [x,y]
        2D array for blue chennel

    Returns
    -------
    rgb_img: ndarray [x,y]
        RGB image

    """
    r_img = np.asarray(r_img, dtype='uint8')
    r_norm_img = (r_img - np.min(r_img)) / (np.max(r_img) - np.min(r_img)).astype(np.uint8)
    
    g_img = np.asarray(g_img, dtype='uint8')
    g_norm_img = (g_img - np.min(g_img)) / (np.max(g_img) - np.min(g_img)).astype(np.uint8)
    
    b_img = np.asarray(b_img, dtype='uint8')
    b_norm_img = (b_img - np.min(b_img)) / (np.max(b_img) - np.min(b_img)).astype(np.uint8)
    
    rgb_img = np.stack([r_norm_img, g_norm_img, b_norm_img], axis=-1) 
    return rgb_img


def arr_cascade_plot(input_arr:np.ndarray, y_shift:float=0.1):  # IN PROGRESS
    plt.figure(figsize=(20, 8))
    
    shift = 0
    for num_ROI in range(input_arr.shape[0]):
        prof_ROI = input_arr[num_ROI]
        plt.plot(prof_ROI+shift, alpha=.5, label=f'ROI {num_ROI}')
        shift += y_shift

    # for line_name in line_dict:
    #     line_lim = line_dict[line_name]
    #     plt.plot(line_lim, [-0.4] * len(line_lim), label=line_name, linewidth=4)

    # plt.vlines(x=[-20], ymin=[-0.1], ymax=[0.0], linewidth=3, color='k')
    # plt.text(x=-30, y=-0.2, s="0.1", size=15, rotation=90.)

    # plt.hlines(y=[-0.75], xmin=[-2], xmax=[8], linewidth=3, color='k')
    # plt.text(x=30, y=-1.15, s="10 s", size=15)

    plt.axis('off')
    plt.legend(loc=2)
    plt.show()


def stat_line_plot(arr_list: list,
                   lab_list:list,
                   t_scale:int=2,
                   stat_method:str='se',
                   app_bar_dict:dict()=None, show_app_bars:bool=False,
                   stim_t:list[int]=None, show_stim:bool=False,
                   figsize:tuple=(15,10), x_lab:str='NA', y_lab:str='NA', plot_title:str='NA',
                   save_path=False):
    """ Line plot with variance whiskers for set of arrays `[t,val],
    could take as input results of `util.masking.label_prof_arr()` function

    Parameters
    ----------
    arr_list: list
        list of 2D arrays with profiles `[t, val]`
    lab_list: list
        list of labels for lines, should be same len with arr_list
    t_scale: float
        time scale for x-axis, in seconds between two frames
    stat_method: str, optional {'se', 'iqr', 'ci'}
        method for line plot calculation: `se` - mean +/- standat error of mean,
        `iqr` - median +/- IQR, `ci` - mean +/- 95% confidence interval
    """
    time_line = np.linspace(0, arr_list[0].shape[1]*t_scale, \
                            num=arr_list[0].shape[1])

    # https://aegis4048.github.io/comprehensive_confidence_intervals_for_python_developers

    # mean, se
    arr_se_stat = lambda x: (np.mean(x, axis=0), \
                             np.std(x, axis=0)/np.sqrt(x.shape[1]))  # mean, se
    
    # meadian, IQR
    arr_iqr_stat = lambda x: (np.median(x, axis=0), \
                              stats.iqr(x, axis=0))

    # mean, CI
    arr_ci_stat = lambda x, alpha=0.05: (np.mean(x, axis=0), \
                                         stats.t.ppf(1-alpha/2, df=x.shape[1]-1) \
                                                    *np.std(x, axis=0, ddof=1)/np.sqrt(x.shape[1]))

    stat_dict = {'se':arr_se_stat,
                 'iqr':arr_iqr_stat,
                 'ci':arr_ci_stat}
    max_list = []
    plt.figure(figsize=figsize)
    for num in range(len(arr_list)):
        arr = arr_list[num]
        lab = lab_list[num]
        arr_val, arr_var = stat_dict[stat_method](arr)
        max_list.append(arr_val.max())

        plt.errorbar(time_line, arr_val,
                     yerr = arr_var,
                     fmt ='-o', capsize=2, label=lab)

    if show_stim:
        for s_t in stim_t:
            plt.axvline(x=s_t, color='k', linestyle='--', linewidth=1.5)

    if show_app_bars:
            for bar_name in app_bar_dict:
                bar_lim = app_bar_dict[bar_name]
                plt.plot(bar_lim, [max(max_list)+0.05] * len(bar_lim),
                         linewidth=4, color='k')

    plt.grid(color='grey', linewidth=.25)
    plt.xlabel(x_lab)
    plt.ylabel(y_lab)
    plt.title(plot_title)
    plt.legend()
    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=300)
    else:
        plt.show()


def overlay_line_plot(input_arr: np.ndarray,
                      t_scale:int=2,
                      min_amp:float=0,max_amp:float=10,
                      app_bar_dict:dict()=None, show_app_bars:bool=False,
                      stim_t:list[int]=None, show_stim:bool=False,
                      figsize:tuple=(15,10), x_lab:str='NA', y_lab:str='NA', plot_title:str='NA',
                      save_path=False):
    """ Line plot with variance whiskers for array `[t,val],
    could take as input results of `util.masking.label_prof_arr()` function.
    Plots every row from the input array as an individual line

    Parameters
    ----------
    input_arr: ndarray [dF_value,t]
        list of 2D arrays with profiles `[t, val]`
    min_amp: float, optional
        minimal amplitude of displayed profiles
    max_amp: float, optional
        maximal amplitude of displayed profiles

    """
    time_line = np.linspace(0, input_arr.shape[1]*t_scale, \
                            num=input_arr.shape[1])

    max_list = []
    plt.figure(figsize=figsize)

    for num_ROI in range(input_arr.shape[0]):
        prof_ROI = input_arr[num_ROI]
        if (prof_ROI.max() < min_amp) | (prof_ROI.max() > max_amp):
            continue
        else:
            plt.plot(time_line, prof_ROI, alpha=.65, marker='o')
            max_list.append(prof_ROI.max())

    if show_stim:
        for s_t in stim_t:
            plt.axvline(x=s_t, color='k', linestyle='--', linewidth=1.5)

    if show_app_bars:
            for bar_name in app_bar_dict:
                bar_lim = app_bar_dict[bar_name]
                plt.plot(bar_lim, [max(max_list)+0.05] * len(bar_lim),
                         linewidth=4, color='k')

    plt.grid(color='grey', linewidth=.25)
    plt.xlabel(x_lab)
    plt.ylabel(y_lab)
    plt.title(f'{plot_title}, min {min_amp}, max {max_amp}')
    plt.legend()
    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=300)
    else:
        plt.show()


import sys
sys.path.append('../src')
from load_save import load_data
from main import analyze_FRET_data

DATA_DIR = r'../example'
EXP_DIR = r'../example/gauss_noise/sinusoids'
	
def run():

	data = load_data(DATA_DIR, 'data_sinusoids')
	a = analyze_FRET_data(data, EXP_DIR)
	a.run()

if __name__ == '__main__':
	run()